/*
 * Line6 Linux USB driver - 0.6.5
 *
 * Copyright (C) 2004-2006 Markus Grabner (grabner@icg.tu-graz.ac.at)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include "config.h"

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/usb.h>

#include "audio.h"
#include "capture.h"
#include "control.h"
#include "driver.h"
#include "midi.h"
#include "playback.h"
#include "pod.h"
#include "usbdefs.h"
#include "variax.h"


#define DRIVER_AUTHOR  "Markus Grabner <grabner@icg.tu-graz.ac.at>"
#define DRIVER_DESC    "Line6 USB Driver"
#define DRIVER_VERSION "0.6.5"


/* table of devices that work with this driver */
static struct usb_device_id line6_id_table [] = {
	{ USB_DEVICE(LINE6_VENDOR_ID, LINE6_DEVID_BASSPODXT) },
	{ USB_DEVICE(LINE6_VENDOR_ID, LINE6_DEVID_BASSPODXTLIVE) },
	{ USB_DEVICE(LINE6_VENDOR_ID, LINE6_DEVID_BASSPODXTPRO) },
	{ USB_DEVICE(LINE6_VENDOR_ID, LINE6_DEVID_PODXT) },
	{ USB_DEVICE(LINE6_VENDOR_ID, LINE6_DEVID_PODXTLIVE) },
	{ USB_DEVICE(LINE6_VENDOR_ID, LINE6_DEVID_PODXTPRO) },
	{ USB_DEVICE(LINE6_VENDOR_ID, LINE6_DEVID_VARIAX) },
	{ USB_DEVICE(LINE6_VENDOR_ID, LINE6_DEVID_TONEPORT) },
	{ },
};
MODULE_DEVICE_TABLE (usb, line6_id_table);

static const char *line6_name_table[] = {
	"BassPODxt",
	"BassPODxt Live",
	"BassPODxt Pro",
	"PODxt",
	"PODxt Live",
	"PODxt Pro",
	"Variax Workbench",
	"TonePort UX1",
	0
};

static const int line6_bits_table[] = {
	LINE6_BIT_BASSPODXT,
	LINE6_BIT_BASSPODXTLIVE,
	LINE6_BIT_BASSPODXTPRO,
	LINE6_BIT_PODXT,
	LINE6_BIT_PODXTLIVE,
	LINE6_BIT_PODXTPRO,
	LINE6_BIT_VARIAX,
	LINE6_BIT_TONEPORT,
	0
};

/* sysex related stuff: */
const unsigned char line6_sysex_header[] = { 0x00, 0x01, 0x0c };

void *line6_devices[LINE6_MAX_DEVICES];


/*
	Forward declarations.
*/
static void line6_data_received(struct urb *urb, struct pt_regs *regs);

/*
	Start to listen on endpoint.
*/
static int line6_start_listen(struct usb_line6 *line6)
{
	usb_fill_bulk_urb(line6->urb_listen,
										line6->usbdev,
										usb_rcvbulkpipe(line6->usbdev, line6->ep_control_read),
										line6->buffer_listen, LINE6_BUFSIZE_LISTEN,
										line6_data_received,
										line6);

	line6->urb_listen->actual_length = 0;
	return usb_submit_urb(line6->urb_listen, GFP_KERNEL);
}

#if DO_URB_DUMP_SEND || DO_URB_DUMP_RECEIVE
/*
	Write hexdump to syslog.
*/
void line6_write_hexdump(struct usb_line6 *line6, char dir, const unsigned char *buffer, int size)
{
	static const int BYTES_PER_LINE = 8;
	char hexdump[100];
	char asc[BYTES_PER_LINE + 1];
	int i, j;

	for(i = 0; i < size; i += BYTES_PER_LINE) {
		int hexdumpsize = sizeof(hexdump);
		char *p = hexdump;
		int n = min(size - i, BYTES_PER_LINE);
		asc[n] = 0;

		for(j = 0; j < BYTES_PER_LINE; ++j) {
			int bytes;

			if(j < n) {
				unsigned char val = buffer[i + j];
				bytes = snprintf(p, hexdumpsize, " %02X", val);
				asc[j] = ((val >= 0x20) && (val < 0x7f)) ? val : '.';
			}
			else
				bytes = snprintf(p, hexdumpsize, "   ");
		
			if(bytes > hexdumpsize)
				break;  /* buffer overflow */

			p += bytes;
			hexdumpsize -= bytes;
		}

		dev_info(line6->msgdev, "%c%04X:%s %s\n", dir, i, hexdump, asc);
	}
}
#endif

#if DO_URB_DUMP_RECEIVE
/*
	Dump URB data to syslog.
*/
static void line6_dump_urb(struct urb *urb)
{
	struct usb_line6 *line6 = (struct usb_line6 *)urb->context;

	if(urb->status < 0)
		return;

	line6_write_hexdump(line6, 'R', (unsigned char *)urb->transfer_buffer, urb->actual_length);
}
#endif

/*
	Send raw message in pieces of 32 bytes.
*/
int line6_send_raw_message(struct usb_line6 *line6, const char *buffer, int size)
{
	int i, done = 0;
	
#if DO_URB_DUMP_SEND
	line6_write_hexdump(line6, 'S', buffer, size);
#endif

	for(i = 0; i < size; i += LINE6_RAW_CHUNK_SIZE)	{
		int partial;
		const char *frag_buf = buffer + i;
		int frag_size = min(LINE6_RAW_CHUNK_SIZE, size - i);
		int retval = usb_bulk_msg(line6->usbdev,
															usb_sndbulkpipe(line6->usbdev, line6->ep_control_write),
															(char *)frag_buf, frag_size, &partial, LINE6_TIMEOUT * HZ);

		if(retval) {
			dev_err(line6->msgdev, "usb_bulk_msg failed (%d)\n", retval);
			break;
		}

		done += frag_size;
	}

	return done;
}

/*
	Notification of completion of asynchronous request transmission.
*/
static void line6_async_request_sent(struct urb *urb, struct pt_regs *regs)
{
	usb_free_urb(urb);
}

/*
	Asynchronously send raw message.
*/
int line6_send_raw_message_async(struct usb_line6 *line6, const char *buffer, int size)
{
	struct urb *urb;
	int retval;

	urb = usb_alloc_urb(0, GFP_ATOMIC);

	if(urb == NULL) {
		dev_err(line6->msgdev, "Out of memory\n");
		return -ENOMEM;
	}

	usb_fill_bulk_urb(urb,
										line6->usbdev,
										usb_sndbulkpipe(line6->usbdev, line6->ep_control_write),
										(char *)buffer, size,
										line6_async_request_sent, line6);

#if DO_URB_DUMP_SEND
	line6_write_hexdump(line6, 'S', buffer, size);
#endif

	retval = usb_submit_urb(urb, GFP_ATOMIC);

	if(retval < 0) {
		dev_err(line6->msgdev, "usb_submit_urb failed 1\n");
		usb_free_urb(urb);
		return -EINVAL;
	}

	return 0;
}

/*
	Send sysex message in pieces of 32 bytes.
*/
int line6_send_sysex_message(struct usb_line6 *line6, const char *buffer, int size)
{
	return line6_send_raw_message(line6, buffer, size + SYSEX_EXTRA_SIZE) - SYSEX_EXTRA_SIZE;
}

/*
	Allocate buffer for sysex message and prepare header.
	@param code sysex message code
	@param size number of bytes between code and sysex end
*/
char *line6_alloc_sysex_buffer(struct usb_line6 *line6, int code1, int code2, int size)
{
	char *buffer = kmalloc(size + SYSEX_EXTRA_SIZE, GFP_KERNEL);

	if(!buffer) {
		dev_err(line6->msgdev, "out of memory\n");
		return 0;
	}

	buffer[0] = LINE6_SYSEX_BEGIN;
	memcpy(buffer + 1, line6_sysex_header, sizeof(line6_sysex_header));
	buffer[sizeof(line6_sysex_header) + 1] = code1;
	buffer[sizeof(line6_sysex_header) + 2] = code2;
	buffer[sizeof(line6_sysex_header) + 3 + size] = LINE6_SYSEX_END;
	return buffer;
}

/*
	Notification of data received from the PODxt Pro.
*/
static void line6_data_received(struct urb *urb, struct pt_regs *regs)
{
	unsigned char *data;
	int length = 0;
	struct usb_line6 *line6 = (struct usb_line6 *)urb->context;

	if(urb->status == -ESHUTDOWN)
		return;

#if DO_URB_DUMP_RECEIVE
	line6_dump_urb(urb);
#endif

	data = urb->transfer_buffer;
	length = urb->actual_length;
	line6_midi_receive(line6, data, length);

	while(length > 0) {
		int bytes = 1, complete = 0;

		if(line6->message_length == 0) {
			if(data[0] < 0x80) {
				DEBUG_MESSAGES(dev_err(line6->msgdev, "message header missed\n"));
				goto finish;
			}
			
			/* determine message length: */
			switch(data[0]) {
			case LINE6_PARAM_CHANGE | LINE6_OFFSET_ORIG:
			case LINE6_PARAM_CHANGE | LINE6_OFFSET_ECHO:
				line6->message_length = 3;
				break;

			case LINE6_PROGRAM_CHANGE | LINE6_OFFSET_ORIG:
			case LINE6_PROGRAM_CHANGE | LINE6_OFFSET_ECHO:
				line6->message_length = 2;
				break;

			case LINE6_SYSEX_BEGIN:
			case LINE6_SYSEX_BEGIN | LINE6_OFFSET_ORIG:
			case LINE6_SYSEX_BEGIN | LINE6_OFFSET_UNKNOWN:
				line6->message_length = -1;
				break;

			case LINE6_RESET:
				line6->message_length = 1;
				break;

			case LINE6_SYSEX_END:
				line6->message_pos = 0;
				line6->message_length = 0;
				goto finish;

			default:
				DEBUG_MESSAGES(dev_err(line6->msgdev, "unknown message %02X\n", (int)data[0]));
				goto finish;
			}
		}

		/* find number of bytes in the URB that contribute to the current message: */
		if(line6->message_length > 0) {
			/* fixed length message: */
			bytes = min(line6->message_length - line6->message_pos, length);

			if(line6->message_pos + bytes >= line6->message_length)
				complete = 1;
		}
		else {
			/* variable length message (sysex): */
			for(bytes = 0; bytes < length; ++bytes)
				if(data[bytes] == LINE6_SYSEX_END) {
					complete = 1;
					++bytes;  /* include sysex end */
					break;
				}
		}

		if(line6->message_pos + bytes > LINE6_MESSAGE_MAXLEN) {
			DEBUG_MESSAGES(dev_err(line6->msgdev, "buffer overflow - message skipped\n"));
			line6->message_pos = 0;
			line6->message_length = 0;
		}
		else {
			memcpy(line6->buffer_message + line6->message_pos, data, length);
			line6->message_pos += bytes;

			if(complete) {
				switch(line6->usbdev->descriptor.idProduct) {
				case LINE6_DEVID_BASSPODXT:
				case LINE6_DEVID_BASSPODXTLIVE:
				case LINE6_DEVID_BASSPODXTPRO:
				case LINE6_DEVID_PODXT:
				case LINE6_DEVID_PODXTLIVE:
				case LINE6_DEVID_PODXTPRO:
					pod_process_message((struct usb_line6_pod *)line6);
					break;
					
				case LINE6_DEVID_VARIAX:
					variax_process_message((struct usb_line6_variax *)line6);
					break;
				}

				line6->message_pos = 0;
				line6->message_length = 0;
			}
		}
	
	finish:	
		data += bytes;
		length -= bytes;
	}

	line6_start_listen(line6);
}

/*
	Send channel number (i.e., switch to a different sound).
*/
int line6_send_program(struct usb_line6 *line6, int value)
{
	int retval;
	unsigned char *buffer;
	unsigned int partial;

	buffer = kmalloc(2, GFP_KERNEL);

	if(!buffer) {
		dev_err(line6->msgdev, "out of memory\n");
		return -ENOMEM;
	}

	buffer[0] = LINE6_PROGRAM_CHANGE;
	buffer[1] = value;

#if DO_URB_DUMP_SEND
	line6_write_hexdump(line6, 'S', buffer, 2);
#endif

	retval = usb_bulk_msg(line6->usbdev,
												usb_sndbulkpipe(line6->usbdev, line6->ep_control_write),
												buffer, 2, &partial, LINE6_TIMEOUT * HZ);

	if(retval)
		dev_err(line6->msgdev, "usb_bulk_msg failed (%d)\n", retval);

	kfree(buffer);
	return retval;
}

/*
	Transmit Line6 control parameter.
*/
int line6_transmit_parameter(struct usb_line6 *line6, int param, int value)
{
	int retval;
	unsigned char *buffer;
	unsigned int partial;

	buffer = kmalloc(3, GFP_KERNEL);

	if(!buffer) {
		dev_err(line6->msgdev, "out of memory\n");
		return -ENOMEM;
	}

	buffer[0] = LINE6_PARAM_CHANGE;
	buffer[1] = param;
	buffer[2] = value;

#if DO_URB_DUMP_SEND
	line6_write_hexdump(line6, 'S', buffer, 3);
#endif

	retval = usb_bulk_msg(line6->usbdev,
												usb_sndbulkpipe(line6->usbdev, line6->ep_control_write),
												buffer, 3, &partial, LINE6_TIMEOUT * HZ);

	if(retval)
		dev_err(line6->msgdev, "usb_bulk_msg failed (%d)\n", retval);

	kfree(buffer);
	return retval;
}

/*
	No operation (i.e., unsupported).
*/
ssize_t line6_nop_read(struct device *dev, DEVICE_ATTRIBUTE char *buf)
{
	return 0;
}

/*
	No operation (i.e., unsupported).
*/
ssize_t line6_nop_write(struct device *dev, DEVICE_ATTRIBUTE const char *buf, size_t count)
{
	return count;
}

/*
	"write" request on "raw" special file.
*/
#if CREATE_RAW_FILE
ssize_t line6_set_raw(struct device *dev, DEVICE_ATTRIBUTE const char *buf, size_t count)
{
	struct usb_interface *interface = to_usb_interface(dev);
	struct usb_line6 *line6 = usb_get_intfdata(interface);
	line6_send_raw_message(line6, buf, count);
	return count;
}
#endif

/*
	Generic destructor.
*/
static void line6_destruct(struct usb_interface *interface)
{
	struct usb_line6 *line6 = usb_get_intfdata(interface);

	if(line6 == NULL)
		return;

	/* free buffer memory first: */
	if(line6->buffer_message) kfree(line6->buffer_message);
	if(line6->buffer_listen) kfree(line6->buffer_listen);

	/* then free URBs: */
	if(line6->urb_listen) usb_free_urb(line6->urb_listen);
}

static void line6_list_devices(void)
{
	int i;

	for(i = 0; i < LINE6_MAX_DEVICES; ++i) {
		struct usb_line6 *dev = line6_devices[i];
		printk(KERN_INFO "Line6 device %d: ", i);

		if(dev == 0)
			printk("(not used)\n");
		else
			printk("%s\n", dev->devname);
	}
}

/*
	Probe USB device.
*/
static int line6_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	int devtype;
	const char *devname;
	struct usb_device *usbdev = interface_to_usbdev(interface);
	struct usb_line6 *line6 = 0;
	int devnum;
	int alternate;
	int product;
	int size = sizeof(struct usb_line6);
	int ep_read = 0, ep_write = 0;
	int ret;

	/* we don't handle multiple configurations */
	if(usbdev->descriptor.bNumConfigurations != 1)
		return -ENODEV;

	/* check vendor and product id */
	for(devtype = sizeof(line6_id_table) / sizeof(line6_id_table[0]) - 1; devtype--;)
		if((le16_to_cpu(usbdev->descriptor.idVendor) == line6_id_table[devtype].idVendor) &&
			 (le16_to_cpu(usbdev->descriptor.idProduct) == line6_id_table[devtype].idProduct))
			break;

	if(devtype < 0)
		return -ENODEV;

	/* check interface number (PODxt Live has interface 1 - what's that?) */
	if(interface->cur_altsetting->desc.bInterfaceNumber != 0)
		return -ENODEV;

	/* find free slot in device table: */
	for(devnum = 0; devnum < LINE6_MAX_DEVICES; ++devnum)
		if(line6_devices[devnum] == 0)
			break;

	if(devnum == LINE6_MAX_DEVICES)
		return -ENODEV;

	/* initialize device info: */
	devname = line6_name_table[devtype];
	dev_info(&interface->dev, "Line6 %s found\n", devname);

	if(usb_reset_configuration(usbdev) < 0) {
		dev_err(&interface->dev, "reset_configuration failed\n");
		return -EINVAL;
	}

	product = le16_to_cpu(usbdev->descriptor.idProduct);

	switch(product) {
	case LINE6_DEVID_PODXTLIVE:
	case LINE6_DEVID_BASSPODXTLIVE:
	case LINE6_DEVID_VARIAX:
		alternate = 1;
		break;

	default:
		alternate = 5;
	}

	if(usb_set_interface(usbdev, LINE6_INTERFACE, alternate) < 0) {
		dev_err(&interface->dev, "set_interface failed\n");
		return -EINVAL;
	}

	/* initialize device data: */
	switch(product) {
	case LINE6_DEVID_BASSPODXT:
	case LINE6_DEVID_BASSPODXTLIVE:
	case LINE6_DEVID_BASSPODXTPRO:
	case LINE6_DEVID_PODXT:
	case LINE6_DEVID_PODXTLIVE:
	case LINE6_DEVID_PODXTPRO:
		size = sizeof(struct usb_line6_pod);
		ep_read = POD_EP_READ_MIDI;
		ep_write = POD_EP_WRITE_MIDI;
		break;

	case LINE6_DEVID_VARIAX:
		size = sizeof(struct usb_line6_variax);
		ep_read = VARIAX_EP_READ_MIDI;
		ep_write = VARIAX_EP_WRITE_MIDI;
		break;
	}

	line6 = kmalloc(size, GFP_KERNEL);

	if(line6 == NULL) {
		dev_err(&interface->dev, "Out of memory\n");
		return -ENOMEM;
	}

	memset(line6, 0, size);

	/* store basic data: */
	line6->devname = devname;
	line6->devbit = line6_bits_table[devtype];
	line6->usbdev = usb_get_dev(usbdev);
	line6->msgdev = &interface->dev;
	usb_set_intfdata(interface, line6);
	line6_devices[devnum] = line6;  /* should be only after successful initialization! */
	line6_list_devices();

	/* initialize USB buffers: */
	line6->ep_control_read = ep_read;
	line6->ep_control_write = ep_write;
	line6->buffer_listen = kmalloc(LINE6_BUFSIZE_LISTEN, GFP_KERNEL);

	if(line6->buffer_listen == NULL) {
		dev_err(&interface->dev, "Out of memory\n");
		line6_destruct(interface);
		return -ENOMEM;
	}

	line6->buffer_message = kmalloc(LINE6_MESSAGE_MAXLEN, GFP_KERNEL);

	if(line6->buffer_message == NULL) {
		dev_err(&interface->dev, "Out of memory\n");
		line6_destruct(interface);
		return -ENOMEM;
	}

	line6->urb_listen = usb_alloc_urb(0, GFP_KERNEL);

	if(line6->urb_listen == NULL) {
		dev_err(&interface->dev, "Out of memory\n");
		line6_destruct(interface);
		return -ENOMEM;
	}

	if(line6_start_listen(line6) < 0) {
		dev_err(&interface->dev, "usb_submit_urb failed 2\n");
		line6_destruct(interface);
		return -EINVAL;
	}

	switch(product) {
	case LINE6_DEVID_BASSPODXT:
	case LINE6_DEVID_BASSPODXTLIVE:
	case LINE6_DEVID_BASSPODXTPRO:
	case LINE6_DEVID_PODXT:
	case LINE6_DEVID_PODXTLIVE:
	case LINE6_DEVID_PODXTPRO:
		ret = pod_init(interface, (struct usb_line6_pod *)line6);
		break;

	case LINE6_DEVID_VARIAX:
		ret = variax_init(interface, (struct usb_line6_variax *)line6);
		break;

	default:
		dev_err(&interface->dev, "Line6 product %04x not supported\n", product);
		ret = -ENODEV;
	}

	if(ret == 0)
		dev_info(&interface->dev, "Line6 %s now attached\n", line6->devname);

	return ret;
}

/*
	Line6 device disconnected.
*/
static void line6_disconnect(struct usb_interface *interface)
{
	struct usb_line6 *line6 = usb_get_intfdata(interface);
	int i;

	usb_kill_urb(line6->urb_listen);
	line6_destruct(interface);

	switch(line6->usbdev->descriptor.idProduct) {
	case LINE6_DEVID_BASSPODXT:
	case LINE6_DEVID_BASSPODXTLIVE:
	case LINE6_DEVID_BASSPODXTPRO:
	case LINE6_DEVID_PODXT:
	case LINE6_DEVID_PODXTLIVE:
	case LINE6_DEVID_PODXTPRO:
		pod_disconnect(interface);
		break;

	case LINE6_DEVID_VARIAX:
		variax_disconnect(interface);
		break;
	}

	usb_put_dev(line6->usbdev);
	dev_info(&interface->dev, "Line6 %s now disconnected\n", line6->devname);
	kfree(line6);
	usb_set_intfdata(interface, NULL);  /* make sure the device isn't destructed twice */

	for(i = LINE6_MAX_DEVICES; i--;)
		if(line6_devices[i] == line6)
			line6_devices[i] = 0;

	line6_list_devices();
}

static struct usb_driver line6_driver = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 16)
	.owner =	THIS_MODULE,
#endif
	.name =		DRIVER_NAME,
	.probe =	line6_probe,
	.disconnect =	line6_disconnect,
	.id_table =	line6_id_table,
};

/*
	Module initialization.
*/
static int __init line6_init(void)
{
	int i, retval;

	for(i = LINE6_MAX_DEVICES; i--;)
		line6_devices[i] = 0;

	retval = usb_register(&line6_driver);

	if(retval)
		err("usb_register failed. Error number %d", retval);

	return retval;
}

/*
	Module cleanup.
*/
static void __exit line6_exit(void)
{
	usb_deregister(&line6_driver);
}

module_init(line6_init);
module_exit(line6_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
MODULE_VERSION(DRIVER_VERSION);
