/*
 * Line6 Linux USB driver - 0.6.5
 *
 * Copyright (C) 2004-2006 Markus Grabner (grabner@icg.tu-graz.ac.at)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include "config.h"

#include <linux/usb.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/rawmidi.h>

#include "audio.h"
#include "driver.h"
#include "midi.h"
#include "pod.h"
#include "usbdefs.h"


#define line6_rawmidi_substream_midi(substream) ((snd_line6_midi_t *)((substream)->rmidi->private_data))


static int send_midi_async(struct usb_line6 *line6, unsigned char *data, int length);


/*
	Pass data received via USB to MIDI.
*/
void line6_midi_receive(struct usb_line6 *line6, unsigned char *data, int length)
{
	if(line6->midi->substream_receive)
		snd_rawmidi_receive(line6->midi->substream_receive, data, length);
}

/*
	Read data from MIDI buffer and transmit them via USB.
*/
static void line6_midi_transmit(snd_rawmidi_substream_t *substream)
{
	unsigned long flags;
	unsigned char data[LINE6_RAW_CHUNK_SIZE];
	struct usb_line6 *line6 = line6_rawmidi_substream_midi(substream)->line6;
	spin_lock_irqsave(&line6->midi->midi_transmit_lock, flags);

	for(;;) {
		int done = snd_rawmidi_transmit_peek(substream, data, sizeof(data));
		int err = 0;

		if(done == 0)
			break;

		if(!((done == 1) && (data[0] == 0xfe)))  /* skip "Active Sense" messages */
			err = send_midi_async(line6, data, done);

		if(err == 0)
			snd_rawmidi_transmit_ack(substream, done);

		if(done != sizeof(data))
			break;
	}

	spin_unlock_irqrestore(&line6->midi->midi_transmit_lock, flags);
}

/*
	Notification of completion of MIDI transmission.
*/
static void midi_sent(struct urb *urb, struct pt_regs *regs)
{
	unsigned long flags;
	int status;
	int num;
	struct usb_line6 *line6 = (struct usb_line6 *)urb->context;

	status = urb->status;
	kfree(urb->transfer_buffer);
	usb_free_urb(urb);

	if(status == -ESHUTDOWN)
		return;

	spin_lock_irqsave(&line6->midi->send_urb_lock, flags);
	num = --line6->midi->num_active_send_urbs;

	if(num == 0) {
		line6_midi_transmit(line6->midi->substream_transmit);
		num = line6->midi->num_active_send_urbs;
	}

	if(num == 0)
		wake_up_interruptible(&line6->midi->send_wait);

	spin_unlock_irqrestore(&line6->midi->send_urb_lock, flags);
}

/*
	Send an asynchronous MIDI message.
	Assumes that line6->midi->send_urb_lock is held
	(i.e., this function is serialized).
*/
static int send_midi_async(struct usb_line6 *line6, unsigned char *data, int length)
{
	struct urb *urb;
	int retval;
	unsigned char *transfer_buffer;

	urb = usb_alloc_urb(0, GFP_ATOMIC);

	if(urb == 0) {
		dev_err(line6->msgdev, "Out of memory\n");
		return -ENOMEM;
	}

#if DO_URB_DUMP_SEND
	line6_write_hexdump(line6, 'S', data, length);
#endif

	transfer_buffer = (unsigned char *)kmalloc(length, GFP_ATOMIC);

	if(transfer_buffer == 0) {
		usb_free_urb(urb);
		dev_err(line6->msgdev, "Out of memory\n");
		return -ENOMEM;
	}

	memcpy(transfer_buffer, data, length);
	usb_fill_bulk_urb(urb,
										line6->usbdev,
										usb_sndbulkpipe(line6->usbdev, line6->ep_control_write),
										transfer_buffer, length, midi_sent, line6);
	urb->actual_length = 0;
	retval = usb_submit_urb(urb, GFP_ATOMIC);

	if(retval < 0) {
		dev_err(line6->msgdev, "usb_submit_urb failed\n");
		usb_free_urb(urb);
		return -EINVAL;
	}

	++line6->midi->num_active_send_urbs;

	switch(line6->usbdev->descriptor.idProduct) {
	case LINE6_DEVID_BASSPODXT:
	case LINE6_DEVID_BASSPODXTLIVE:
	case LINE6_DEVID_BASSPODXTPRO:
	case LINE6_DEVID_PODXT:
	case LINE6_DEVID_PODXTLIVE:
	case LINE6_DEVID_PODXTPRO:
		pod_midi_postprocess((struct usb_line6_pod *)line6, data, length);
		break;
	}

	return 0;
}

static int line6_midi_output_open(snd_rawmidi_substream_t *substream)
{
	return 0;
}

static int line6_midi_output_close(snd_rawmidi_substream_t *substream)
{
	return 0;
}

static void line6_midi_output_trigger(snd_rawmidi_substream_t *substream, int up)
{
	unsigned long flags;
	struct usb_line6 *line6 = line6_rawmidi_substream_midi(substream)->line6;

	line6->midi->substream_transmit = substream;
	spin_lock_irqsave(&line6->midi->send_urb_lock, flags);

	if(line6->midi->num_active_send_urbs == 0)
		line6_midi_transmit(substream);

	spin_unlock_irqrestore(&line6->midi->send_urb_lock, flags);
}

static void line6_midi_output_drain(snd_rawmidi_substream_t *substream)
{
	static int count = 0;
	struct usb_line6 *line6 = line6_rawmidi_substream_midi(substream)->line6;
	wait_queue_head_t *head = &line6->midi->send_wait;
	DECLARE_WAITQUEUE(wait, current);
	add_wait_queue(head, &wait);
	current->state = TASK_INTERRUPTIBLE;

	while(line6->midi->num_active_send_urbs > 0)
		if(signal_pending(current))
			break;
		else
			schedule();

	current->state = TASK_RUNNING;
	remove_wait_queue(head, &wait);
	++count;
}

static int line6_midi_input_open(snd_rawmidi_substream_t *substream)
{
	return 0;
}

static int line6_midi_input_close(snd_rawmidi_substream_t *substream)
{
	return 0;
}

static void line6_midi_input_trigger(snd_rawmidi_substream_t *substream, int up)
{
	struct usb_line6 *line6 = line6_rawmidi_substream_midi(substream)->line6;

	if(up)
		line6->midi->substream_receive = substream;
	else
		line6->midi->substream_receive = 0;
}

static snd_rawmidi_ops_t line6_midi_output_ops = {
	.open = line6_midi_output_open,
	.close = line6_midi_output_close,
	.trigger = line6_midi_output_trigger,
	.drain = line6_midi_output_drain,
};

static snd_rawmidi_ops_t line6_midi_input_ops = {
	.open = line6_midi_input_open,
	.close = line6_midi_input_close,
	.trigger = line6_midi_input_trigger,
};

/*
	Cleanup the Line6 MIDI device.
*/
static void line6_cleanup_midi(snd_rawmidi_t *rmidi)
{
}

/* create a MIDI device */
static int snd_line6_new_midi(struct snd_line6_midi *midi)
{
	snd_rawmidi_t *rmidi;
	int err;

	if((err = snd_rawmidi_new(midi->line6->card, "Line6 MIDI", 0, 1, 1, &rmidi)) < 0)
		return err;

	rmidi->private_data = midi;
	rmidi->private_free = line6_cleanup_midi;
	strcpy(rmidi->name, midi->line6->devname);

	rmidi->info_flags =
		SNDRV_RAWMIDI_INFO_OUTPUT |
		SNDRV_RAWMIDI_INFO_INPUT |
		SNDRV_RAWMIDI_INFO_DUPLEX;

	snd_rawmidi_set_ops(rmidi, SNDRV_RAWMIDI_STREAM_OUTPUT, &line6_midi_output_ops);
	snd_rawmidi_set_ops(rmidi, SNDRV_RAWMIDI_STREAM_INPUT, &line6_midi_input_ops);
	return 0;
}

/* MIDI device destructor */
static int snd_line6_midi_free(snd_device_t *device)
{
	return 0;
}

/*
	Initialize the Line6 MIDI subsystem.
*/
int line6_init_midi(struct usb_line6 *line6)
{
	static snd_device_ops_t midi_ops = {
		.dev_free = snd_line6_midi_free,
	};

	int err;
	snd_line6_midi_t *midi;

	midi = kmalloc(sizeof(snd_line6_midi_t), GFP_KERNEL);

	if(midi == NULL)
		return -ENOMEM;

	memset(midi, 0, sizeof(snd_line6_midi_t));
	midi->line6 = line6;
	line6->midi = midi;

	if((err = snd_device_new(line6->card, SNDRV_DEV_RAWMIDI, line6, &midi_ops)) < 0)
		return err;

	snd_card_set_dev(line6->card, line6->msgdev);

	if((err = snd_line6_new_midi(midi)) < 0)
		return err;

	init_waitqueue_head(&midi->send_wait);
	spin_lock_init(&midi->send_urb_lock);
	spin_lock_init(&midi->midi_transmit_lock);
	return 0;
}
