/*
 * Line6 Linux USB driver - 0.6.5
 *
 * Copyright (C) 2004-2006 Markus Grabner (grabner@icg.tu-graz.ac.at)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#include "config.h"

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/control.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>

#include "audio.h"
#include "capture.h"
#include "driver.h"
#include "playback.h"
#include "pod.h"


/* trigger callback */
int snd_pod_trigger(snd_pcm_substream_t *substream, int cmd)
{
	snd_pod_pcm_t *chip = snd_pcm_substream_chip(substream);
	struct list_head *pos;
	snd_pcm_substream_t *s;
	int err;
	unsigned long flags;

	spin_lock_irqsave(&chip->lock_trigger, flags);
	clear_bit(BIT_PREPARED, &chip->flags);

	snd_pcm_group_for_each(pos, substream) {
		s = snd_pcm_group_substream_entry(pos);

		switch(s->stream) {
		case SNDRV_PCM_STREAM_PLAYBACK:
			err = snd_pod_playback_trigger(s, cmd);

			if(err < 0) {
				spin_unlock_irqrestore(&chip->lock_trigger, flags);
				return err;
			}

			break;

		case SNDRV_PCM_STREAM_CAPTURE:
			err = snd_pod_capture_trigger(s, cmd);

			if(err < 0) {
				spin_unlock_irqrestore(&chip->lock_trigger, flags);
				return err;
			}

			break;

		default:
			dev_err(s2m(substream), "Unknown stream direction %d\n", s->stream);
		}
	}

	spin_unlock_irqrestore(&chip->lock_trigger, flags);
	return 0;
}

/* control info callback */
static int snd_pod_control_info(snd_kcontrol_t *kcontrol, snd_ctl_elem_info_t *uinfo) {
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 2;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = 256;
	return 0;
}

/* control get callback */
static int snd_pod_control_get(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
	int i;
	snd_pod_pcm_t *chip = snd_kcontrol_chip(kcontrol);

	for(i = 2; i--;)
		ucontrol->value.integer.value[i] = chip->volume[i];

	return 0;
}

/* control put callback */
static int snd_pod_control_put(snd_kcontrol_t *kcontrol, snd_ctl_elem_value_t *ucontrol) {
	int i, changed = 0;
	snd_pod_pcm_t *chip = snd_kcontrol_chip(kcontrol);

	for(i = 2; i--;)
		if(chip->volume[i] != ucontrol->value.integer.value[i]) {
			chip->volume[i] = ucontrol->value.integer.value[i];
			changed = 1;
		}

	return changed;
}

/* control definition */
static snd_kcontrol_new_t pod_control = {
	.iface = SNDRV_CTL_ELEM_IFACE_MIXER,
	.name = "PCM Playback Volume",
	.index = 0,
	.access = SNDRV_CTL_ELEM_ACCESS_READWRITE,
	.info = snd_pod_control_info,
	.get = snd_pod_control_get,
	.put = snd_pod_control_put
};

/*
	Cleanup the PODxt Pro PCM device.
*/
static void pod_cleanup_pcm(snd_pcm_t *pcm)
{
	int i;
	struct snd_pod_pcm *chip = snd_pcm_chip(pcm);

	for(i = POD_ISO_BUFFERS; i--;) {
		if(chip->urb_audio_out[i]) {
			usb_kill_urb(chip->urb_audio_out[i]);
			usb_free_urb(chip->urb_audio_out[i]);
		}
		if(chip->urb_audio_in[i]) {
			usb_kill_urb(chip->urb_audio_in[i]);
			usb_free_urb(chip->urb_audio_in[i]);
		}
	}
}

/* create a PCM device */
static int snd_pod_new_pcm(struct snd_pod_pcm *chip)
{
	snd_pcm_t *pcm;
	int err;

	if((err = snd_pcm_new(chip->pod->line6.card, (char *)chip->pod->line6.devname, 0, 1, 1, &pcm)) < 0) 
		return err;

	pcm->private_data = chip;
	pcm->private_free = pod_cleanup_pcm;
	chip->pcm = pcm;
	strcpy(pcm->name, chip->pod->line6.devname);

	/* set operators */
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &snd_pod_playback_ops);
	snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &snd_pod_capture_ops);

	/* pre-allocation of buffers */
	snd_pcm_lib_preallocate_pages_for_all(pcm,
																				SNDRV_DMA_TYPE_CONTINUOUS,
																				snd_dma_continuous_data(GFP_KERNEL),
																				64 * 1024, 128 * 1024);

	return 0;
}

/* PCM device destructor */
static int snd_pod_pcm_free(snd_device_t *device)
{
	return 0;
}

/*
	Create and register the PCM device and mixer entries.
	Create URBs for playback and capture.
*/
int pod_init_pcm(struct usb_line6_pod *pod)
{
	static snd_device_ops_t pcm_ops = {
		.dev_free = snd_pod_pcm_free,
	};

	int err;
	snd_pod_pcm_t *chip;
	chip = kmalloc(sizeof(snd_pod_pcm_t), GFP_KERNEL);

	if(chip == NULL)
		return -ENOMEM;

	memset(chip, 0, sizeof(snd_pod_pcm_t));
	chip->volume[0] = chip->volume[1] = 128;
	chip->pod = pod;
	pod->chip = chip;

	/* PCM device: */
	if((err = snd_device_new(pod->line6.card, SNDRV_DEV_PCM, pod, &pcm_ops)) < 0)
		return err;

	snd_card_set_dev(pod->line6.card, pod->line6.msgdev);

	if((err = snd_pod_new_pcm(chip)) < 0)
		return err;

	spin_lock_init(&chip->lock_audio_out);
	spin_lock_init(&chip->lock_audio_in);
	spin_lock_init(&chip->lock_trigger);

	if((err = create_audio_out_urbs(chip)) < 0)
		return err;

	if((err = create_audio_in_urbs(chip)) < 0)
		return err;

	/* mixer: */
	if((err = snd_ctl_add(pod->line6.card, snd_ctl_new1(&pod_control, chip))) < 0)
		return err;

	return 0;
}

/* prepare pcm callback */
int snd_pod_prepare(snd_pcm_substream_t *substream)
{
	snd_pod_pcm_t *chip = snd_pcm_substream_chip(substream);

	if(!test_and_set_bit(BIT_PREPARED, &chip->flags)) {
		unlink_wait_clear_audio_out_urbs(chip);
		chip->pos_out = 0;
		chip->pos_out_done = 0;
	
		unlink_wait_clear_audio_in_urbs(chip);
		chip->bytes_out = 0;
		chip->pos_in_done = 0;
		chip->bytes_in = 0;
	}

	return 0;
}
