/*
 * Line6 Linux USB driver - 0.6.5
 *
 * Copyright (C) 2004-2006 Markus Grabner (grabner@icg.tu-graz.ac.at)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 */

#ifndef POD_H
#define POD_H


#include <linux/spinlock.h>
#include <linux/usb.h>
#include <linux/wait.h>

#include <sound/driver.h>
#include <sound/core.h>

#include "driver.h"
#include "dumprequest.h"


/*
	POD endpoints
*/
#define POD_EP_WRITE_AUDIO 0x01
#define POD_EP_WRITE_MIDI  0x03
#define POD_EP_READ_AUDIO  0x82
#define POD_EP_READ_MIDI   0x84

/*
	Locate name in binary program dump
*/
#define	POD_NAME_OFFSET 0
#define	POD_NAME_LENGTH 16

/*
	Other constants
*/
#define POD_CONTROL_SIZE 0x80
#define POD_BUFSIZE_DUMPREQ 7


/**
	 Data structure for values that need to be requested explicitly.
	 This is the case for system and tuner settings.
*/
struct ValueWait
{
	unsigned short value;
	wait_queue_head_t wait;
};

/**
	 Binary PodXT Pro program dump
*/
struct pod_program {
	/**
		 Header information (including program name).
	*/
	unsigned char header[0x20];

	/**
		 Program parameters.
	*/
	unsigned char control[POD_CONTROL_SIZE];
};

struct usb_line6_pod {
	/**
		 Generic Line6 USB data.
	*/
	struct usb_line6 line6;

	/**
		 Dump request structure.
	*/
	struct line6_dump_request dumpreq;

	/**
		 PODxt Pro PCM device data structure.
	*/
	struct snd_pod_pcm *chip;

	/**
		 Current program number.
	*/
	unsigned char channel_num;

	/**
		 Current program settings.
	*/
	struct pod_program prog_data;

	/**
		 Buffer for data retrieved from or to be stored on PODxt Pro.
	*/
	struct pod_program prog_data_buf;

#if USE_CHARACTER_DEVICE
	/**
		 Status of character device;
		 <0:  not registered
		 >=0: successfully registered
	*/
	int chrdev;
#endif

	/**
		 Buffer for requesting version number.
	*/
	unsigned char *buffer_versionreq;

	/**
		 Wait queue for access to batch data.
	*/
	wait_queue_head_t batch_wait;

	/**
		 Tuner mute mode.
	*/
	struct ValueWait tuner_mute;

	/**
		 Tuner base frequency (typically 440Hz).
	*/
	struct ValueWait tuner_freq;

	/**
		 Note received from tuner.
	*/
	struct ValueWait tuner_note;

	/**
		 Pitch value received from tuner.
	*/
	struct ValueWait tuner_pitch;

	/**
		 Instrument monitor level.
	*/
	struct ValueWait monitor_level;

	/**
		 Audio routing mode.
		 0: send processed guitar
		 1: send clean guitar
		 2: send clean guitar re-amp playback
		 3: send re-amp playback
	*/
	struct ValueWait routing;

	/**
		 Dirty flags for access to parameter data.
	*/
	unsigned long param_dirty[POD_CONTROL_SIZE / sizeof(unsigned long)];

#if USE_CHARACTER_DEVICE
	/**
		 Buffer for reading from character device.
	*/
	unsigned char chrdev_readbuf[9];

	/**
		 Buffer for writing to character device.
	*/
	unsigned char chrdev_writebuf[3];

	/**
		 Length of current message;
	*/
	short chrdev_msglen;

	/**
		 Current position in character device read buffer.
	*/
	short chrdev_readpos;

	/**
		 Current position in character device write buffer.
	*/
	short chrdev_writepos;
#endif

	/**
		 Some atomic flags.
	*/
	unsigned long atomic_flags;

	/**
		 Serial number of device.
	*/
	int serial_number;

	/**
		 Firmware version (x 100).
	*/
	int firmware_version;

	/**
		 Device ID.
	*/
	int device_id;

	/**
		 Flag to indicate modification of current program settings.
	*/
	char dirty;

	/**
		 Flag if initial firmware version request has been successful.
	*/
	char versionreq_ok;

	/**
		 Flag to enable MIDI postprocessing.
	*/
	char midi_postprocess;
};


extern void pod_disconnect(struct usb_interface *interface);
extern int pod_init(struct usb_interface *interface, struct usb_line6_pod *pod);
extern void pod_midi_postprocess(struct usb_line6_pod *pod, unsigned char *data, int length);
extern void pod_process_message(struct usb_line6_pod *pod);
extern void pod_receive_parameter(struct usb_line6_pod *pod, int param);
extern void pod_transmit_parameter(struct usb_line6_pod *pod, int param, int value);


#endif
